#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Common GUI elements."""

import sys
from io import BytesIO
from os import path

import matplotlib.backends.qt_editor.figureoptions
from matplotlib import pyplot
from matplotlib._pylab_helpers import Gcf
from matplotlib.backend_bases import key_press_handler
from matplotlib.backends.backend_qt5agg import \
    FigureCanvasQTAgg as FigureCanvas, FigureManagerQT, NavigationToolbar2QT
from matplotlib.colors import LogNorm
from matplotlib.figure import Figure

from ufit.qt import QByteArray, QDialog, QFileDialog, QIcon, QLineEdit, \
    QMessageBox, QPageLayout, QPainter, QPrintDialog, QPrinter, \
    QPrintPreviewWidget, QRectF, QSettings, QSize, QSizePolicy, QSvgRenderer, \
    Qt, pyqtSignal, uic

pyplot.rc('font', family='sans-serif')
pyplot.rc('font', **{'sans-serif': 'Sans Serif, Arial, Helvetica, '
                     'Lucida Grande, DejaVu Sans'})

# pylint: disable=wrong-import-position
from ufit.gui import logger
from ufit.gui.ploteditor import figure_edit
from ufit.gui.session import session
from ufit.plotting import DataPlotter

# override figure editor with our extended version
matplotlib.backends.qt_editor.figureoptions.figure_edit = figure_edit

uipath = path.dirname(__file__)


def loadUi(widget, uiname, subdir='ui'):
    uic.loadUi(path.join(uipath, subdir, uiname), widget)


def path_to_str(qstring):
    return qstring


def str_to_path(string):
    if not isinstance(string, str):
        return string.decode(sys.getfilesystemencoding())
    return string


class MPLCanvas(FigureCanvas):
    """Ultimately, this is a QWidget (as well as a FigureCanvasAgg, etc.)."""

    logzChanged = pyqtSignal()
    replotRequest = pyqtSignal()

    def __init__(self, parent, width=10, height=6, dpi=72, maincanvas=False):
        fig = Figure(figsize=(width, height), dpi=dpi)
        fig.set_facecolor('white')
        self.print_width = 0
        self.main = parent
        self.logz = False
        self.axes = fig.add_subplot(111)
        self.plotter = DataPlotter(self, self.axes)
        # make tight_layout do the right thing
        self.axes.set_xlabel('x')
        self.axes.set_ylabel('y')
        self.axes.set_title('(data title)\n(info)')
        FigureCanvas.__init__(self, fig)

        # create a figure manager so that we can use pylab commands on the
        # main viewport
        def make_active(event):
            Gcf.set_active(self.manager)
        self.manager = FigureManagerQT(self, 1)
        self.manager._cidgcf = self.mpl_connect('button_press_event', make_active)
        Gcf.set_active(self.manager)
        self.setParent(parent)
        self.setSizePolicy(QSizePolicy.Policy.Expanding,
                           QSizePolicy.Policy.Expanding)
        self.updateGeometry()
        # actually get key events
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.mpl_connect('key_press_event', self.key_press)
        # These will not do anything in standalone mode, but do not hurt.
        if maincanvas:
            session.propsRequested.connect(self.on_session_propsRequested)
            session.propsUpdated.connect(self.on_session_propsUpdated)

    def on_session_propsRequested(self):
        session.props.canvas_logz = self.logz

    def on_session_propsUpdated(self):
        if 'canvas_logz' in session.props:
            self.logz = session.props.canvas_logz
            self.logzChanged.emit()

    def key_press(self, event):
        key_press_handler(event, self)

    def resizeEvent(self, event):
        FigureCanvas.resizeEvent(self, event)
        self.figure.tight_layout(pad=2)

    def print_(self):
        sio = BytesIO()
        self.print_figure(sio, format='svg')
        svg = QSvgRenderer(QByteArray(sio.getvalue()))
        sz = svg.defaultSize()
        aspect = sz.width()/float(sz.height())

        printer = QPrinter(QPrinter.PrinterMode.HighResolution)
        printer.setPageOrientation(QPageLayout.Orientation.Landscape)

        dlg = QDialog(self)
        loadUi(dlg, 'printpreview.ui')
        dlg.width.setValue(self.print_width or 500)
        ppw = QPrintPreviewWidget(printer, dlg)
        dlg.layout().insertWidget(1, ppw)

        def render(printer):
            height = printer.height() * (dlg.width.value()/1000.)
            width = aspect * height
            painter = QPainter(printer)
            svg.render(painter, QRectF(0, 0, width, height))

        def sliderchanged(newval):
            ppw.updatePreview()

        ppw.paintRequested.connect(render)
        dlg.width.valueChanged.connect(sliderchanged)
        if dlg.exec() != QDialog.DialogCode.Accepted:
            return
        self.print_width = dlg.width.value()
        pdlg = QPrintDialog(printer, self)
        if pdlg.exec() != QDialog.DialogCode.Accepted:
            return
        render(printer)

    def ufit_replot(self):
        self.replotRequest.emit()


class MPLToolbar(NavigationToolbar2QT):

    popoutRequested = pyqtSignal()

    icon_name_map = {
        'home.png':         'magnifier-zoom-fit.png',
        'back.png':         'arrow-180.png',
        'forward.png':      'arrow.png',
        'move.png':         'arrow-move.png',
        'zoom_to_rect.png': 'selection-resize.png',
        'filesave.png':     'document-pdf.png',
        'printer.png':      'printer.png',
        'pyconsole.png':    'terminal--arrow.png',
        'log-x.png':        'log-x.png',
        'log-y.png':        'log-y.png',
        'log-z.png':        'log-z.png',
        'exwindow.png':    'chart--arrow.png',
    }

    toolitems = list(NavigationToolbar2QT.toolitems)
    del toolitems[7]  # subplot adjust
    toolitems.insert(0, ('Log x', 'Logarithmic X scale', 'log-x',
                         'logx_callback'))
    toolitems.insert(1, ('Log y', 'Logarithmic Y scale', 'log-y',
                         'logy_callback'))
    toolitems.insert(2, ('Log z', 'Logarithmic Z scale for images', 'log-z',
                         'logz_callback'))
    toolitems.insert(3, (None, None, None, None))
    toolitems.append(('Print', 'Print the figure', 'printer',
                      'print_callback'))
    toolitems.append(('Pop out', 'Show the figure in a separate window',
                      'exwindow', 'popout_callback'))
    toolitems.append(('Execute', 'Show Python console', 'pyconsole',
                      'exec_callback'))

    def __init__(self, *args, **kwds):
        NavigationToolbar2QT.__init__(self, *args, **kwds)
        self.locLabel.setAlignment(Qt.AlignmentFlag.AlignRight |
                                   Qt.AlignmentFlag.AlignVCenter)
        self._actions['logx_callback'].setCheckable(True)
        self._actions['logy_callback'].setCheckable(True)
        self._actions['logz_callback'].setCheckable(True)
        self.canvas.logzChanged.connect(self.on_canvas_logzChanged)

    def _icon(self, name, color=None):
        if name in self.icon_name_map:
            return QIcon(':/' + self.icon_name_map[name])
        return QIcon()

    def home(self, *args):
        # always unzoom completely
        if hasattr(self, '_views'):
            self._views.clear()
        if hasattr(self, '_positions'):
            self._positions.clear()
        self.canvas.figure.gca().autoscale()
        self.canvas.draw()
        return NavigationToolbar2QT.home(self)

    def logx_callback(self):
        ax = self.canvas.figure.gca()
        if ax.get_xscale() == 'linear':
            ax.set_xscale('symlog')
            self._actions['logx_callback'].setChecked(True)
        else:
            ax.set_xscale('linear')
            self._actions['logx_callback'].setChecked(False)
        self.canvas.draw()

    def logy_callback(self):
        ax = self.canvas.figure.gca()
        if ax.get_yscale() == 'linear':
            ax.set_yscale('symlog')
            self._actions['logy_callback'].setChecked(True)
        else:
            ax.set_yscale('linear')
            self._actions['logy_callback'].setChecked(False)
        self.canvas.draw()

    def logz_callback(self):
        ax = self.canvas.figure.gca()
        self.canvas.logz = not self.canvas.logz
        session.set_dirty()
        self._actions['logz_callback'].setChecked(self.canvas.logz)
        for im in ax.get_images():
            if self.canvas.logz:
                im.set_norm(LogNorm())
            else:
                im.set_norm(None)
        self.canvas.draw()

    def on_canvas_logzChanged(self):
        self._actions['logz_callback'].setChecked(self.canvas.logz)

    def print_callback(self):
        self.canvas.print_()

    def popout_callback(self):
        self.popoutRequested.emit()

    def exec_callback(self):
        try:
            from ufit.gui.console import ConsoleWindow
        except ImportError:
            logger.exception('Qt console window cannot be opened without '
                             'IPython; import error was:')
            QMessageBox.information(self, 'ufit',
                                    'Please install IPython with qtconsole to '
                                    'activate this function.')
            return
        try:
            w = ConsoleWindow(self)
        except Exception as err:
            logger.exception('Qt console window cannot be opened without '
                             'IPython; import error was:')
            QMessageBox.information(self, 'ufit', 'IPython qtconsole '
                                    f'seems to be broken: {err}')
            return
        w.ipython.executeCommand('from ufit.lab import *')
        w.ipython.pushVariables({
            'fig': self.canvas.figure,
            'ax': self.canvas.figure.gca(),
            'D': [item for group in session.groups for item in group.items],
        })
        w.show()

    def save_figure(self, *args):
        filetypes = self.canvas.get_supported_filetypes_grouped()
        sorted_filetypes = sorted(filetypes.items())

        start = self.canvas.get_default_filename()
        filters = []
        for name, exts in sorted_filetypes:
            if 'eps' in exts or 'emf' in exts or 'jpg' in exts or \
               'pgf' in exts or 'raw' in exts:
                continue
            exts_list = " ".join(['*.%s' % ext for ext in exts])
            filt = '%s (%s)' % (name, exts_list)
            filters.append(filt)
        filters = ';;'.join(filters)
        fname, _ = QFileDialog.getSaveFileName(
            self, 'Choose a filename to save to', start, filters)
        if fname:
            try:
                self.canvas.print_figure(str(fname))
            except Exception as e:
                logger.exception('Error saving file')
                QMessageBox.critical(self, 'Error saving file', str(e))


class SmallLineEdit(QLineEdit):
    def sizeHint(self):
        sz = QLineEdit.sizeHint(self)
        return QSize(round(sz.width() / 1.5), sz.height())


class SettingGroup:
    def __init__(self, name):
        self.name = name
        self.settings = QSettings('ufit', 'gui')

    def __enter__(self):
        self.settings.beginGroup(self.name)
        return self.settings

    def __exit__(self, *args):
        self.settings.endGroup()
        self.settings.sync()



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for PSI ZEBRA data files."""

import io

from numpy import array, concatenate, loadtxt, zeros


def guess_cols(colnames, coldata, meta):
    if 'hkle_vary' in meta:
        xcol = {'h': 'H', 'k': 'K', 'l': 'L', 'E': 'NP'}[meta['hkle_vary']]
    else:
        xcol = colnames[1]
    return xcol, 'Counts', None, 'Monitor1'


def check_data(fp):
    line = fp.readline()
    fp.seek(0, 0)
    return line.startswith(b'##SICS ASCII at ZEBRA')


def read_data(filename, fp):
    fp = io.TextIOWrapper(fp, 'ascii', 'ignore')

    meta = {}
    try:
        meta['filenumber'] = int(filename.split('n')[-1].split('.')[0])
    except Exception:
        meta['filenumber'] = 0
    meta['instrument'] = 'ZEBRA'
    remark = ''
    title = ''

    # parse headers
    lines = iter(fp)
    for line in lines:
        line = line.strip()
        if line == '#data':
            break
        if not line or line.startswith('#'):
            continue
        key, val = line.split('=', 1)
        key = key.strip()
        oval = val.strip()
        try:
            val = float(oval)
        except ValueError:
            val = oval
        if key == 'title':
            title = oval
        elif key == 'ProposalID':
            meta['experiment'] = oval
        elif key == 'sample':
            remark = oval
        elif key == 'comment':
            meta['subtitle'] = oval
        elif key == 'original_filename':
            if meta['filenumber'] == 0:
                try:
                    meta['filenumber'] = \
                        int(filename.split('n')[-1].split('.')[0])
                except Exception:
                    pass
        else:
            meta[key] = val

    if remark:
        title += ', ' + remark
    meta['title'] = title

    # parse data table
    next(lines)   # Scanning variables
    parts = next(lines).split()   # npoints, mease. mode
    meta['mode'] = parts[3].strip(',').lower()
    meta['preset'] = float(parts[5])
    colnames = next(lines).split()
    arr = loadtxt(lines, comments='E')  # E as in END-OF-DATA

    # if number of colnames is not correct, discard them
    if len(colnames) != arr.shape[1]:
        colnames = ['Column %d' % i for i in range(1, arr.shape[1]+1)]

    cols = dict((name, arr[:, i]) for (i, name) in enumerate(colnames))
    meta['environment'] = []
    if 'H' in colnames:
        npoints = arr.shape[0]
        hkle = concatenate([arr[:, (1, 2, 3)], zeros((npoints, 1))], axis=1)
        meta['hkle'] = hkle
        deviations = array([(cs.max()-cs.min()) for cs in meta['hkle'].T])
        meta['hkle_vary'] = ['h', 'k', 'l', 'E'][deviations.argmax()]
    for col in cols:
        meta[col] = cols[col].mean()
    if 'temp' in cols:
        if meta['temp'] > 0:
            meta['environment'].append('T = %.3f K' % meta['temp'])
    if 'mf' in cols:
        if meta['mf'] > 0:
            meta['environment'].append('B = %.3f T' % meta['mf'])

    return colnames, arr, meta



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for DESY scattering data."""

import io
import time

from numpy import array, genfromtxt


def check_data(fp):
    line1 = fp.readline()
    line2 = fp.readline()
    fp.seek(0, 0)
    return line1.strip() == b'!' and line2.strip() == b'! Comments'


def guess_cols(colnames, coldata, meta):
    xg, yg, mg = None, 'mag_c06', 'sumvfcs_counts'
    if 'h_position' in colnames:
        qhindex = colnames.index('h_position')
        deviations = array([(cs.max()-cs.min())
                            for cs in coldata.T[qhindex:qhindex+3]])
        xg = colnames[qhindex + deviations.argmax()]
    return xg, yg, None, mg


def read_data(filename, fp):
    fp = io.TextIOWrapper(fp, 'ascii', 'ignore')
    meta = {}
    mode = None
    cline = 0
    names = []

    for line in fp:
        if line.startswith('!'):
            pass
        elif line.startswith('%'):
            mode = line[1:2]
        elif mode == 'c':  # Comments
            if cline == 0:
                meta['subtitle'] = line.strip()
            elif cline == 1:
                meta['user'] = line.split()[1]
                ix = line.find('Acquisition started')
                if ix > 0:
                    meta['created'] = time.mktime(time.strptime(
                        line[ix:].strip(),
                        'Acquisition started at %a %b %d %H:%M:%S %Y'))
            cline += 1
        elif mode == 'p':  # Parameters
            try:
                name, val = line.split(' = ', 1)
                meta[name] = float(val)
            except ValueError:
                pass
        elif mode == 'd':  # Data
            if line.startswith(' Col'):
                parts = line.split()
                names.append(parts[2])
            else:
                lines = [line]
                lines.extend(fp)
                arr = genfromtxt(lines, comments='!')
                break

    for i, n in enumerate(names):
        meta[n] = arr[:, i].mean()

    meta['environment'] = []
    for name in names:
        if name.endswith('persistentfieldt'):
            meta['environment'].append('B = %.5f T' % meta[name])
        elif name.endswith('tempprobe'):
            meta['environment'].append('T = %.3f K' % meta[name])

    if 'h_position' in names:
        qhindex = names.index('h_position')
        meta['hkle'] = arr[:, qhindex:qhindex+3]
        meta['hkle'] = array([(h, k, l, 0) for (h, k, l) in meta['hkle']])
        deviations = array([(cs.max()-cs.min())
                            for cs in arr.T[qhindex:qhindex+3]])
        meta['hkle_vary'] = ['h', 'k', 'l', 'E'][deviations.argmax()]

    return names, arr, meta



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Embedded IPython qt console."""

# pylint: disable=wrong-import-order
from qtconsole.inprocess import QtInProcessKernelManager
from qtconsole.rich_jupyter_widget import RichJupyterWidget

from ufit.qt import QMainWindow, pyqtSignal


class QIPythonWidget(RichJupyterWidget):  # pylint: disable=abstract-method
    """Convenience class for a live IPython console widget."""

    closeme = pyqtSignal()
    redrawme = pyqtSignal()

    # pylint: disable=keyword-arg-before-vararg
    def __init__(self, customBanner=None, *args, **kwargs):
        if customBanner is not None:
            self.banner = customBanner
            self.kernel_banner = ''
        super().__init__(*args, **kwargs)
        self.kernel_manager = kernel_manager = QtInProcessKernelManager()
        kernel_manager.start_kernel()
        kernel_manager.kernel.gui = 'qt'
        self.kernel_client = kernel_client = self._kernel_manager.client()
        kernel_client.start_channels()

        def stop():
            kernel_client.stop_channels()
            kernel_manager.shutdown_kernel()
            self.closeme.emit()
        self.exit_requested.connect(stop)

        def redraw():
            self.redrawme.emit()
        self.executed.connect(redraw)

    def pushVariables(self, var_dict):
        """Given a dictionary containing name / value pairs, push those
        variables to the IPython console widget.
        """
        self.kernel_manager.kernel.shell.push(var_dict)

    def clearTerminal(self):
        """Clears the terminal."""
        self._control.clear()

    def printText(self, text):
        """Prints some plain text to the console."""
        self._append_plain_text(text)

    def executeCommand(self, command):
        """Execute a command in the frame of the console widget."""
        self._execute(command, True)


class ConsoleWindow(QMainWindow):
    def __init__(self, parent=None):
        QMainWindow.__init__(self, parent)
        self.setWindowTitle('ufit Python console')
        self.ipython = QIPythonWidget(
            '''\
ufit interactive Python shell

Objects in the namespace:
* fig -- figure of the main viewport
* ax  -- current axes of the main viewport
* D   -- list of dataset items
            ''',
            self)
        self.setCentralWidget(self.ipython)

        self.ipython.closeme.connect(self.close)
        self.ipython.redrawme.connect(parent.canvas.draw)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

# Original author: Douglas Creager <dcreager@dcreager.net>
# This file is placed into the public domain.

import os.path
from subprocess import PIPE, Popen

__all__ = ['get_version']

RELEASE_VERSION_FILE = os.path.join(os.path.dirname(__file__),
                                    'RELEASE-VERSION')
GIT_REPO = os.path.join(os.path.dirname(__file__), '..', '.git')


def translate_version(ver):
    ver = ver.lstrip('v').rsplit('-', 2)
    return f'{ver[0]}.post{ver[1]}+{ver[2]}' if len(ver) == 3 else ver[0]


def get_git_version(abbrev=4):
    try:
        with Popen(['git', f'--git-dir={GIT_REPO}',
                    'describe', f'--abbrev={abbrev}'],
                   stdout=PIPE, stderr=PIPE) as p:
            stdout, _stderr = p.communicate()
        return translate_version(stdout.strip().decode('utf-8', 'ignore'))
    except Exception:
        return None


def read_release_version():
    try:
        with open(RELEASE_VERSION_FILE, encoding='utf-8') as f:
            return f.readline().strip()
    except Exception:
        return None


def write_release_version(version):
    with open(RELEASE_VERSION_FILE, 'w', encoding='utf-8') as f:
        f.write(f'{version}\n')


def get_version(abbrev=4):
    # determine the version from git and from RELEASE-VERSION
    git_version = get_git_version(abbrev)
    release_version = read_release_version()

    # if we have a git version, it is authoritative
    if git_version:
        if git_version != release_version:
            write_release_version(git_version)
        return git_version
    if release_version:
        return release_version
    raise ValueError('Cannot find a version number - make sure that '
                     'git is installed or a RELEASE-VERSION file is '
                     'present!')


if __name__ == "__main__":
    print(get_version())



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for DynaCool PPMS data files."""

import io
from os import path

from numpy import genfromtxt, mean, ptp


def guess_cols(colnames, coldata, meta):
    xcol = 'Time Stamp (sec)'
    for cn in colnames:
        if cn.replace(' ', '').lower().startswith('timestamp'):
            xcol = cn
    ycol = dycol = None
    if 'AC Moment (emu)' in colnames:
        ycol = 'AC Moment (emu)'
        dycol = 'AC Std. Dev. (emu)'
    elif 'Moment (emu)' in colnames:
        ycol = 'Moment (emu)'
        dycol = 'M. Std. Err. (emu)'
    elif 'Samp HC (µJ/K)' in colnames:
        ycol = 'Samp HC (µJ/K)'
        dycol = 'Samp HC Err (µJ/K)'
    elif 'Bridge 1 Resistivity (Ohm-m)' in colnames:
        ycol = 'Bridge 1 Resistivity (Ohm-m)'
        dycol = 'Bridge 1 Std. Dev. (Ohm-m)'
    else:
        # fallback
        ycol = [colnames[i] for i in range(len(colnames))
                if ptp(coldata[:, i]) > 0][-1]
    xcands = ['Magnetic Field (Oe)', 'Temperature (K)', 'Time Stamp (sec)',
              'Time Stamp (Seconds)', 'Sample Temp (Kelvin)']
    spreads = []
    for xcand in xcands:
        if xcand in colnames:
            data = coldata[:, colnames.index(xcand)]
            spreads.append((ptp(data)/mean(data), xcand))
    if spreads:
        xcol = sorted(spreads)[-1][1]
    return xcol, ycol, dycol, None


def check_data(fp):
    fp.readline()
    try:
        line2 = fp.readline()
        line3 = fp.readline()
        line4 = fp.readline()
        fp.seek(0, 0)
        return line2.startswith((b'; VSM Data File',
                                 b'; ACMS II Data File')) or \
               line3.startswith(b'BYAPP,') or \
               line4.startswith(b'BYAPP,')
    except IOError:
        return False


def read_data(filename, fp):
    fp = io.TextIOWrapper(fp, 'latin1', 'ignore')

    meta = {}
    meta['filenumber'] = 0
    try:
        meta['filenumber'] = int(filename.split('_')[-1].split('.')[0])
    except Exception:
        pass
    meta['instrument'] = 'Dynacool'
    meta['filedesc'] = path.basename(filename)
    remark = ''
    title = ''
    is_heatcap = False

    # parse headers
    lines = iter(fp)
    for line in lines:
        line = line.strip()
        if line == '[Data]':
            break
        if not line or line.startswith(';'):
            continue
        parts = line.split(',', 2)
        if parts[0] == 'TITLE':
            title = parts[1]
            continue
        if parts[0] == 'BYAPP' and parts[1] == 'HeatCapacity':
            is_heatcap = True
            continue
        if parts[0] != 'INFO' or len(parts) < 3:
            continue

        if parts[2] == 'VSM_SERIAL_NUMBER':
            meta['instrument'] = parts[1]
        elif parts[2] == 'ACMS_SERIAL_NUMBER':
            meta['instrument'] = parts[1]
        elif parts[2] == 'SAMPLE_MATERIAL':
            remark = parts[1]
        elif parts[2] == 'SAMPLE_COMMENT':
            meta['subtitle'] = parts[1]
        try:
            val = float(parts[1])
        except ValueError:
            val = parts[1]
        meta[parts[2]] = val

    if remark:
        title += ', ' + remark
    meta['title'] = title

    # parse data table
    colnames = next(lines).split(',')
    arr = genfromtxt(lines, delimiter=',', missing_values='0',
                     invalid_raise=False,
                     usecols=list(range(len(colnames))))

    cols = dict((name, arr[:, i]) for (i, name) in enumerate(colnames))
    meta['environment'] = []
    for col in cols:
        meta[col] = cols[col].mean()

    return colnames, arr, meta



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for ILL TAS data."""

import io
import re
from os import path
from warnings import catch_warnings

from numpy import array, atleast_2d, concatenate, cos, dot, genfromtxt, \
    linalg, radians, sin

from ufit import UFitError


def check_data(fp):
    dtline = fp.readline()
    fp.seek(0, 0)
    return dtline.startswith(b'RRRRRRRRRRRR')


def guess_cols(colnames, coldata, meta):
    xg, yg, mg = None, None, None
    if 'CNTS' in colnames:
        yg = 'CNTS'
    if colnames[0] == 'QH':
        if 'EN' in colnames:
            deviations = array([(cs.max()-cs.min()) for cs in coldata.T[:4]])
        else:
            deviations = array([(cs.max()-cs.min()) for cs in coldata.T[:3]])
        xg = colnames[deviations.argmax()]
    else:
        xg = colnames[0]
    maxmon = 0
    for i, colname in enumerate(colnames):
        if colname.startswith('M') and colname[1:].isdigit():
            if coldata[:, i].sum() > maxmon:
                maxmon = coldata[:, i].sum()
                mg = colname
    return xg, yg, None, mg


def read_data(filename, fp):
    fp = io.TextIOWrapper(fp, 'ascii', 'ignore')
    line = ''
    meta = {}
    while line.strip() != 'DATA_:':
        if line.startswith('IIIIIIIIIIIIIIIIIII'):
            # D23 format
            fp.seek(0, 0)
            return read_data_d23(filename, fp)
        if line.startswith('COMND:'):
            meta['subtitle'] = ' '.join(line[7:].rstrip().lower().split())
        elif line.startswith('TITLE:'):
            meta['title'] = line[6:].strip()
        elif line.startswith('FILE_:'):
            meta['filenumber'] = int(line[6:].strip())
        elif line.startswith('PARAM:'):
            parts = line[6:].strip().rstrip(',').split(',')
            for part in parts:
                k, s = part.split('=')
                meta[k.strip()] = float(s.strip())
        elif line.startswith('INSTR:'):
            meta['instrument'] = line[6:].strip().lower()
        elif line.startswith('EXPNO:'):
            meta['experiment'] = line[6:].strip().lower()
        line = fp.readline()
        if not line:
            break
    all_names = fp.readline().split()
    if not all_names:
        raise UFitError('No data columns found in in file %r' % filename)
    usecols = []
    names = []
    for i, name in enumerate(all_names):
        # XXX have to do flipper handling right
        if name in ('PNT', 'F1', 'F2'):
            continue
        names.append(name)
        usecols.append(i)
    # Berlin implementation adds "Finished ..." in the last line,
    # pretend that it is a comment
    with catch_warnings(record=True) as warnings:
        arr = atleast_2d(genfromtxt(iter(lambda: fp.readline().encode(), b''),
                                    usecols=usecols, comments='F',
                                    invalid_raise=False))
    for warning in warnings:
        print('!!! %s' % warning.message)
    for i, n in enumerate(names):
        meta[n] = arr[:, i].mean()
    meta['environment'] = []
    if 'TT' in meta:
        meta['environment'].append('T = %.3f K' % meta['TT'])
    if 'MAG' in meta:
        meta['environment'].append('B = %.5f T' % meta['MAG'])
    if names[3] == 'EN':
        meta['hkle'] = arr[:, :4]
        deviations = array([(cs.max()-cs.min()) for cs in arr.T[:4]])
        meta['hkle_vary'] = ['h', 'k', 'l', 'E'][deviations.argmax()]
    elif names[0] == 'QH':  # 2-axis mode
        meta['hkle'] = arr[:, :3]
        meta['hkle'] = array([(h, k, l, 0) for (h, k, l) in meta['hkle']])
        deviations = array([(cs.max()-cs.min()) for cs in arr.T[:3]])
        meta['hkle_vary'] = ['h', 'k', 'l', 'E'][deviations.argmax()]
    if len(arr) == 0:
        raise UFitError('No data found in file %r' % filename)
    return names, arr, meta


# format of D23 data; wildly impractical
def read_data_d23(filename, fp):
    line = fp.readline()
    meta = {'instrument': 'D23'}
    imeta = {}
    arr = []
    after_head = False
    scantype = None
    section = 0
    while True:
        if len(line) == 81 and len(set(line)) == 2:
            # heading
            after_head = True
            section += 1
        elif after_head:
            if section == 1:
                meta['filenumber'] = int(line.split()[0])
            elif section == 2:
                # experiment user
                fp.readline()
                line = fp.readline()
                meta['experiment'] = line.strip('\x00').split()[0]
            elif section == 3:
                # experiment title and "scan type" (can be empty)
                fp.readline()
                line = fp.readline()
                meta['title'] = line[:72].strip()
                if len(line) >= 72:
                    scantype = line[72:].strip()
            elif section == 4:
                # integral metadata, only relevant for interpreting the rest
                nlines = int(line.split()[1])
                valnames = []
                values = []
                for _ in range(nlines):
                    valnames.extend(fp.readline().split())
                for _ in range(nlines):
                    values.extend(map(int, fp.readline().split()))
                imeta = dict(zip(valnames, values))
            elif section == 5:
                # experimental metadata, take all
                nlines = int(line.split()[1])
                valnames = []
                values = []
                for _ in range(nlines):
                    valnames.extend(x for x in re.split(
                        ' {2,}', fp.readline().strip()) if x)
                for _ in range(nlines):
                    values.extend(map(float, fp.readline().split()))
                for (k, v) in zip(valnames, values):
                    meta[k] = v
            elif section == 7:
                # data follows; must be #points * (#detvals + #angles)
                ncols = 3 + imeta['nbang']
                if int(line.split()[0]) != imeta['npdone'] * ncols:
                    raise ValueError('invalid number of data items')
                line = fp.readline()
                data = []
                while line:
                    data.extend(map(float, line.split()))
                    line = fp.readline()
                arr = array(data).reshape((imeta['npdone'], ncols))
                break
            after_head = False
        line = fp.readline()
        if not line:
            break

    # check if we have a supplementary .dat file to get the HKL values out
    # (the original file doesn't contain it, and we don't want to implement
    # the UB matrix to calculate it here)
    hkl = None
    if scantype != 'omega' and path.isfile(filename + '.dat'):
        hkl = []
        with io.open(filename + '.dat', 'rb') as fp1:
            fp1 = io.TextIOWrapper(fp1, 'ascii', 'ignore')
            for line in fp1:
                hkl.append([float(x) for x in line.split()[:3]])
        hkl = array(hkl)
        meta['hkle'] = array([(h, k, l, 0) for (h, k, l) in hkl])
        deviations = array([(cs.max()-cs.min()) for cs in arr.T[:4]])
        meta['hkle_vary'] = ['h', 'k', 'l', 'E'][deviations.argmax()]

    # otherwise, calculate from UB matrix
    elif scantype != 'omega' and ncols >= 6:
        if 'ub(1,1)' in meta:
            ub = array([[meta['ub(1,1)'], meta['ub(1,2)'], meta['ub(1,3)']],
                        [meta['ub(2,1)'], meta['ub(2,2)'], meta['ub(2,3)']],
                        [meta['ub(3,1)'], meta['ub(3,2)'], meta['ub(3,3)']]])
        else:
            ub = array([[meta['ub(1 1)'], meta['ub(1 2)'], meta['ub(1 3)']],
                        [meta['ub(2 1)'], meta['ub(2 2)'], meta['ub(2 3)']],
                        [meta['ub(3 1)'], meta['ub(3 2)'], meta['ub(3 3)']]])
        assert imeta['icdesc1'] == 1  # gamma
        assert imeta['icdesc2'] in (2, 16)  # omega
        assert imeta['icdesc3'] == 5  # chi
        hkl = []
        for point in arr:
            gamma, omega, nu = radians(point[3:6])
            omatrix = array([[sin(omega), -cos(omega), 0],
                             [cos(omega), sin(omega), 0],
                             [0, 0, 1]])
            gvec = array([cos(gamma)*cos(nu) - 1, sin(gamma)*cos(nu), sin(nu)])
            hkl.append(dot(linalg.inv(ub), dot(omatrix, gvec)))
        hkl = array(hkl) / meta['wavelength']
        meta['hkle'] = array([(h, k, l, 0) for (h, k, l) in hkl])
        deviations = array([(cs.max()-cs.min()) for cs in arr.T[:4]])
        meta['hkle_vary'] = ['h', 'k', 'l', 'E'][deviations.argmax()]

    # reshuffle columns to place the detector last
    arrindices = array(list(range(3, arr.shape[1])) + [1, 2, 0])
    arr = arr[:, arrindices]

    if hkl is not None:
        names = ['QH', 'QK', 'QL']
        arr = concatenate([hkl, arr], 1)
    else:
        names = []
    for i in range(1, 8):
        coltype = imeta['icdesc%d' % i]
        if coltype == 0:
            break
        names.append({
            1: 'GAMMA',
            2: 'OMEGA',
           16: 'OMEGA',
            5: 'CHI',
            -1: 'T',
            -6: 'B',
        }.get(coltype, 'COL_%d' % coltype))
    names.extend(['M1', 'M2', 'CNTS'])
    for i, n in enumerate(names):
        meta[n] = arr[:, i].mean()
    meta['environment'] = []
    if 'Temp-sample' in meta:
        meta['environment'].append('T = %.3f K' % meta['Temp-sample'])
    if 'Mag.field' in meta:
        meta['environment'].append('B = %.5f T' % meta['Mag.field'])

    if len(arr) == 0:
        raise UFitError('No data found in file %r' % filename)
    return names, arr, meta



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

import sys
from os import path

import numpy as np

from ufit.qt import QApplication, QFileDialog, QMainWindow, QVBoxLayout, uic

import ufit.bzplot as bp
import ufit.qreader as qr
from ufit.gui.common import MPLCanvas, MPLToolbar, SettingGroup, path_to_str
from ufit.utils import extract_template


# show the points
class ReciprocalViewer(QMainWindow):

    def __init__(self, parent):
        """ Constructing a basic QApplication
        """
        QMainWindow.__init__(self, parent)
        self.sgroup = SettingGroup('main')
        self.ui = uic.loadUi(path.join(path.dirname(__file__), 'ui', 'qexplorer.ui'), self)
        self.addWidgets()
        self.ui.show()
        self.dir = path.dirname(__file__)
        self.pts = []
        self.canvas = MPLCanvas(self)
        self.canvas.plotter.lines = True
        self.toolbar = MPLToolbar(self.canvas, self)
        self.toolbar.setObjectName('browsetoolbar')
        self.addToolBar(self.toolbar)
        self.v1 = [1, 0, 0]
        self.v2 = [0, 0, 1]
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.canvas)
        self.frmPlot.setLayout(layout)
        self.Reader = None

        # restore settings
        with self.sgroup as settings:
            data_template_path = settings.value('last_data_template', '')
            print(settings.value('test', ''))
            if data_template_path:
                dtempl, numor = extract_template(data_template_path)
                self.dir = dtempl
                self.ui.txtNumors.setText(str(numor))
                print("directory set to", self.dir)

    def closeEvent(self, event):
        with self.sgroup as settings:
            settings.setValue('test', 'abc')

    def addWidgets(self):
        """ Connecting signals """
        self.ui.btnLoad.clicked.connect(self.readData)
        self.ui.btnSelectDir.clicked.connect(self.changeDir)
        self.ui.btnShow.clicked.connect(self.readPoints)
        self.ui.btnAddBZ.clicked.connect(self.addBZ)
        self.ui.fltEmax.valueChanged.connect(self.showPoints)
        self.ui.fltEmin.valueChanged.connect(self.showPoints)
        self.ui.chkBigFont.stateChanged.connect(self.bigFont)

    def readData(self):
        """ This function will read data from files indicated by folder and Numors """

        numors = self.ui.txtNumors.text()
        self.canvas.axes.text(0.5, 0.5, 'Please wait, loading all data...',
                              horizontalalignment='center')
        self.canvas.draw()
        QApplication.processEvents()
        self.reader = qr.QReader(self.dir, numors)
        self.canvas.axes.clear()
        self.canvas.draw()

    def readVectors(self):
        """ Reading orientation vectors from text fields """

        self.v1 = [float(xx) for xx in str(self.ui.txtV1.text()).split()]
        self.v2 = [float(xx) for xx in str(self.ui.txtV2.text()).split()]

    def readPoints(self):
        """ It will find points in read scan files which has hkle values
        These is then converted to orientation vector basis and calculated distance from it.
        Only inplane points are processed.
        """

        self.readVectors()
        self.canvas.axes.text(0.5, 0.5, 'Please wait, parsing read data...',
                              horizontalalignment='center')
        self.canvas.draw()
        QApplication.processEvents()
        self.pts = self.reader.get_points(self.v1, self.v2)
        print("Datafiles read:", len(self.pts))
        self.canvas.axes.clear()
        self.canvas.draw()
        self.showPoints()

    def showPoints(self):
        """ Will show read points in selected E range to canvas  """
        plotter = self.canvas.plotter
        plotter.reset(False)
        x = []
        y = []
        emin = self.ui.fltEmin.value()
        emax = self.ui.fltEmax.value()

        for pt in self.pts:
            if pt[2] >= emin and pt[2] <= emax:  # ok, show point
                x.append(pt[0])
                y.append(pt[1])

        plotter.axes.plot(x, y, 'ro')
        plotter.axes.set_title('Measured points in reciprocal space', size='medium')
        plotter.axes.set_xlabel("x . (%.1f, %.1f, %.1f)" % tuple(self.v1))
        plotter.axes.set_ylabel("y . (%.1f, %.1f, %.1f)" % tuple(self.v2))
        plotter.draw()

    def changeDir(self):
        """ Change directory from which are the data read  """

        if self.dir:
            startdir = self.dir
        else:
            startdir = '.'
        fn = path_to_str(QFileDialog.getOpenFileName(
            self, 'Choose a file', startdir, 'All files (*)')[0])
        if not fn:
            return
        dtempl, numor = extract_template(fn)
        self.dir = dtempl
        self.ui.txtNumors.setText(str(numor))
        print("directory changed to", self.dir)

    def bigFont(self, state):
        """ Increase font size  """
        # TODO: should be done more sofisticated with more configuration

        ax = self.canvas.plotter.axes
        sizes = {
            "normal": [12, 10, 10],
            "big": [24, 22, 18]
        }
        if state == 0:  # normal size
            toset = sizes["normal"]
        else:
            toset = sizes["big"]

        ax.title.set_fontsize(toset[0])
        ax.xaxis.label.set_fontsize(toset[1])
        ax.yaxis.label.set_fontsize(toset[1])
        for t in ax.xaxis.get_major_ticks():
            t.label.set_fontsize(toset[2])
        for t in ax.yaxis.get_major_ticks():
            t.label.set_fontsize(toset[2])
        if ax.legend_:
            for t in ax.legend_.texts:
                t.set_fontsize(toset[1])
        self.canvas.plotter.draw()

    def addBZ(self):
        """
        Experimental: add Brillouin zone.
        Only body centered tetragonal now supported.
        """

        # check by vectors
        self.readVectors()
        myplane = np.cross(self.v1, self.v2)
        # TODO: generate gamma point dynamically
        gpts = np.array([[0, 0], [0, 2], [1, 1], [2, 0], [2, 2]])

        # TODO: read lattice parameters from file
        bzc = bp.BZCreator(gpts, a = 4.33148, c = 10.83387, plane = myplane)
        bzc.doPlot(self.canvas.plotter.axes)
        self.canvas.plotter.draw()


# Run the gui if not imported
if __name__ == "__main__":
    app = QApplication(sys.argv)
    win = ReciprocalViewer()
    sys.exit(app.exec())



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for simple whitespace-separated column data files."""

import io
from os import path

from numpy import insert, loadtxt


def guess_cols(colnames, coldata, meta):
    if len(colnames) > 2:
        dycol = colnames[2]
    else:
        dycol = None
    return colnames[0], colnames[1], dycol, None


def check_data_simple(fp, sep=None):
    line = fp.readline()
    # find the first non-comment line
    while line.startswith((b'#', b'%')):
        line = fp.readline()
    line2 = fp.readline()
    fp.seek(0, 0)
    # special case FRM2 irradiation
    if sep == b',' and line2.startswith(b'Wavelength (nm),Abs,'):
        return True
    # must be values in non-comment line, or the line after
    try:
        [float(x) for x in line.split(sep)]
    except ValueError:
        try:
            [float(x) for x in line2.split(sep)]
        except ValueError:
            return False
    return True


def check_data(fp):
    return check_data_simple(fp, None)


def read_data_simple(filename, fp, sep=None):
    fp = io.TextIOWrapper(fp, 'ascii', 'ignore')
    line1 = ''
    line2 = fp.readline()
    skiprows = 0
    # find the first non-comment line
    while line2.startswith(('#', '%')):
        line1 = line2
        line2 = fp.readline()
        skiprows += 1
    # now line2 is the first non-comment line (but may be column names)

    # check for special case
    line3 = fp.readline()
    if sep == ',' and line3.startswith('Wavelength (nm),Abs,'):
        return read_data_frm2_irradiation(filename, line2, fp)

    # if there are comments, line1 will have the comment char
    comments = '#'
    if line1.startswith(('#', '%')):
        comments = line1[0]
        line1 = line1[1:]
    try:
        [float(x) for x in line2.split(sep)]
    except ValueError:
        # must be column names
        colnames = line2.rstrip().split(sep)
        skiprows += 1
    else:
        # line1 might have column names
        if line1:
            colnames = line1.rstrip().split(sep)
        else:
            colnames = []
    fp.seek(0, 0)
    arr = loadtxt(fp, ndmin=2, skiprows=skiprows, comments=comments,
                  delimiter=sep)
    # if number of colnames is not correct, discard them
    if len(colnames) != arr.shape[1]:
        colnames = ['Column %d' % i for i in range(1, arr.shape[1]+1)]
    meta = {}
    meta['filedesc'] = path.basename(filename)
    return colnames, arr, meta


def read_data(filename, fp):
    return read_data_simple(filename, fp, None)


def read_data_frm2_irradiation(filename, samplenames, fp):
    fp.seek(0, 0)
    cols = samplenames.rstrip().split(',')[:-1]
    colnames = sum((
        [f'{sample} wavelength', f'{sample} abs', f'{sample} abs err']
        for sample in cols[::2]), [])
    usecols = list(range(len(cols)))
    arr = loadtxt(fp, ndmin=2, skiprows=2, delimiter=',', usecols=usecols)
    for i in range(len(cols) // 2):
        arr = insert(arr, 3*i + 2, 0.01, axis=1)
    meta = {'filedesc': path.basename(filename)}
    return colnames, arr, meta



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Fitting/plotting GUI for ufit."""

import argparse
import sys
import time
from os import path

t0 = time.time()

# pylint: disable=wrong-import-position
from ufit.qt import QApplication, QIcon, QMessageBox, QPixmap, QSplashScreen, \
    Qt

from ufit.gui.loggers import getLogger

logger = getLogger('ufit')

from ufit import __version__
from ufit.gui.common import SettingGroup
from ufit.gui.dataloader import start as start_loader
from ufit.gui.fitter import start as start_fitter
from ufit.gui.main import UFitMain
from ufit.utils import extract_template

__all__ = ['start_fitter', 'start_loader']


def main():
    app = QApplication([])
    app.setOrganizationName('ufit')
    app.setApplicationName('gui')
    icon = QIcon()
    icon.addFile(':/appicon.png')
    icon.addFile(':/appicon-16.png')
    icon.addFile(':/appicon-48.png')
    QApplication.setWindowIcon(icon)

    pixmap = QPixmap(':/splash.png')
    splash = QSplashScreen(pixmap, Qt.WindowType.WindowStaysOnTopHint)
    splash.showMessage('Loading...' + '\xa0' * 10 + '\n\n',
                       Qt.AlignmentFlag.AlignRight |
                       Qt.AlignmentFlag.AlignBottom)
    splash.show()
    time.sleep(0.1)
    with SettingGroup('main') as settings:
        if settings.value('current_version', '') != 'v' + __version__:
            settings.setValue('current_version', 'v' + __version__)
            # Execute here the actions to be performed only once after
            # each update (there is nothing there for now, but it could
            # be useful some day...)
            logger.info('Upgrade to version %s finished', __version__)
    app.processEvents()

    def log_unhandled(*exc_info):
        logger.error('Unhandled exception in Qt callback', exc_info=exc_info)
    sys.excepthook = log_unhandled

    t1 = time.time()
    logger.info('Startup: import finished (%.3f s), starting GUI...', t1-t0)

    mainwindow = UFitMain()

    parser = argparse.ArgumentParser(usage='''\
    Usage: %(prog)s [-b directory] [.ufit file | data file]
    ''')
    parser.add_argument('-b', '--browse', action='store', metavar='DIR',
                        help='open browse window in specified directory')
    parser.add_argument('args', nargs=argparse.REMAINDER,
                        help='.ufit or data files')

    opts = parser.parse_args()

    if len(opts.args) >= 1:
        datafile = path.abspath(opts.args[0])
        if path.isdir(datafile):
            # directory given, treat it as -b argument (browse)
            opts.browse = datafile
        elif datafile.endswith('.ufit'):
            try:
                mainwindow.filename = datafile
                mainwindow.load_session(datafile)
            except Exception as err:
                QMessageBox.warning(mainwindow,
                                    'Error', 'Loading failed: %s' % err)
                mainwindow.filename = None
        else:
            dtempl, numor = extract_template(datafile)
            mainwindow.dloader.set_template(dtempl, numor, silent=False)
    mainwindow.show()
    if opts.browse:
        mainwindow.dloader.open_browser(opts.browse)

    t2 = time.time()
    logger.info('Startup: loading finished (%.3f s), main window opened', t2-t1)
    splash.deleteLater()
    app.exec()



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Data operations panel."""

from numpy import arange, linspace, ones
from scipy.fftpack import fft
from scipy.interpolate import interp1d

from ufit.qt import QDialog, QMessageBox, QWidget, pyqtSignal, pyqtSlot

from ufit.data.dataset import ScanData
from ufit.data.merge import floatmerge, rebin
from ufit.gui.common import SettingGroup, loadUi
from ufit.gui.mappingitem import MappingItem
from ufit.gui.session import session


class DataOps(QWidget):
    replotRequest = pyqtSignal(object)
    pickRequest = pyqtSignal(object)
    titleChanged = pyqtSignal()

    def __init__(self, parent):
        QWidget.__init__(self, parent)
        self.item = None
        self.data = None
        self.picking = None
        self.picked_points = []

        loadUi(self, 'dataops.ui')
        self.pickedLbl.hide()

    def initialize(self, item):
        self.data = item.data
        self.model = item.model
        self.item = item
        if self.data.fitmin is not None:
            self.limitminEdit.setText('%.5g' % self.data.fitmin)
        if self.data.fitmax is not None:
            self.limitmaxEdit.setText('%.5g' % self.data.fitmax)
        self.monscaleEdit.setText(str(self.data.nscale))
        self.titleEdit.setText(self.data.title)
        self.nameEdit.setText(self.data.name)
        self.fftNpointsEdit.setText(str(len(self.data.x) * 4))

    def on_canvas_pick(self, event):
        if not hasattr(event, 'artist'):
            return
        if self.picking:
            xdata = event.artist.get_xdata()[event.ind]
            ydata = event.artist.get_ydata()[event.ind]
            self.picked_points.append(xdata)
            self.pickedLbl.setText('%d picked' % len(self.picked_points))
            event.canvas.figure.gca().plot([xdata], [ydata], 'ow', ms=8,
                                           mec='blue')
            event.canvas.draw()

    @pyqtSlot()
    def on_badResetBtn_clicked(self):
        self.data.reset_mask()
        self.replotRequest.emit(None)
        session.set_dirty()

    @pyqtSlot()
    def on_badPointsBtn_clicked(self):
        if self.picking == 'bad':
            self.badPointsBtn.setText('Start')
            self.pickedLbl.hide()
            self.picking = None
            self.removeBadPoints(self.picked_points)
        elif not self.picking:
            self.badPointsBtn.setText('Click points on plot, then '
                                      'here to finish')
            self.pickRequest.emit(self)
            self.picking = 'bad'
            self.picked_points = []
            self.pickedLbl.setText('0 picked')
            self.pickedLbl.show()

    @pyqtSlot()
    def on_limitsBtn_clicked(self):
        try:
            limitmin = float(self.limitminEdit.text())
        except ValueError:
            limitmin = None
        try:
            limitmax = float(self.limitmaxEdit.text())
        except ValueError:
            limitmax = None
        self.data.fitmin, self.data.fitmax = limitmin, limitmax
        self.replotRequest.emit(None)
        session.set_dirty()

    def removeBadPoints(self, points):
        """'Remove' bad data points (just mask them out)."""
        for point in points:
            self.data.mask[self.data.x == point] = False
        self.replotRequest.emit(None)
        session.set_dirty()

    @pyqtSlot()
    def on_rebinBtn_clicked(self):
        try:
            binsize = float(self.precisionEdit.text())
        except ValueError:
            QMessageBox.warning(self, 'Error', 'Enter a valid precision.')
            return
        new_array, new_meta = rebin(self.data._data, binsize, self.data.meta)
        self.data.__init__(new_meta, new_array,
                           self.data.xcol, self.data.ycol, self.data.ncol,
                           self.data.nscale, name=self.data.name,
                           sources=self.data.sources)
        self.replotRequest.emit(None)
        session.set_dirty()

    @pyqtSlot()
    def on_floatmergeBtn_clicked(self):
        try:
            binsize = float(self.precisionEdit.text())
        except ValueError:
            QMessageBox.warning(self, 'Error', 'Enter a valid precision.')
            return
        new_array, new_meta = floatmerge(self.data._data, binsize,
                                         self.data.meta)
        self.data.__init__(new_meta, new_array,
                           self.data.xcol, self.data.ycol, self.data.ncol,
                           self.data.nscale, name=self.data.name,
                           sources=self.data.sources)
        self.replotRequest.emit(None)
        session.set_dirty()

    @pyqtSlot()
    def on_cloneBtn_clicked(self):
        new_data = self.data.copy()
        new_model = self.model.copy()
        from ufit.gui.scanitem import ScanDataItem
        session.add_item(ScanDataItem(new_data, new_model), self.item.group)

    @pyqtSlot()
    def on_mappingBtn_clicked(self):
        item = MappingItem([self.data], None)
        session.add_item(item, self.item.group)

    @pyqtSlot()
    def on_arbBtn_clicked(self):
        dlg = QDialog(self)
        loadUi(dlg, 'change.ui')
        with SettingGroup('main') as settings:
            dlg.xEdit.setText(settings.value('changex', 'x'))
            dlg.yEdit.setText(settings.value('changey', 'y'))
            dlg.dyEdit.setText(settings.value('changedy', 'dy'))
            if dlg.exec() != QDialog.DialogCode.Accepted:
                return
            settings.setValue('changex', dlg.xEdit.text())
            settings.setValue('changey', dlg.yEdit.text())
            settings.setValue('changedy', dlg.dyEdit.text())
        xfml = dlg.xEdit.text()
        yfml = dlg.yEdit.text()
        dyfml = dlg.dyEdit.text()
        new_x = []
        new_y = []
        new_dy = []
        ns = {}
        for dpoint in zip(self.data.x, self.data.y_raw, self.data.dy_raw):
            ns.update(x=dpoint[0], y=dpoint[1], dy=dpoint[2])
            new_x.append(eval(xfml, ns))
            new_y.append(eval(yfml, ns))
            new_dy.append(eval(dyfml, ns))
        self.data.x[:] = new_x
        self.data.y_raw[:] = new_y
        self.data.dy_raw[:] = new_dy
        self.data.y = self.data.y_raw / self.data.norm
        self.data.dy = self.data.dy_raw / self.data.norm
        self.replotRequest.emit(None)
        session.set_dirty()

    @pyqtSlot()
    def on_mulBtn_clicked(self):
        try:
            const = float(self.scaleConstEdit.text())
        except ValueError:
            return
        self.data.y *= const
        self.data.y_raw *= const
        self.data.dy *= const
        self.data.dy_raw *= const
        self.replotRequest.emit(None)
        session.set_dirty()

    @pyqtSlot()
    def on_addBtn_clicked(self):
        try:
            const = float(self.addConstEdit.text())
        except ValueError:
            return
        self.data.y += const
        self.data.y_raw += const * self.data.norm
        self.replotRequest.emit(None)
        session.set_dirty()

    @pyqtSlot()
    def on_scaleXBtn_clicked(self):
        try:
            const = float(self.scaleXConstEdit.text())
        except ValueError:
            return
        self.data.x *= const
        self.replotRequest.emit(None)
        session.set_dirty()

    @pyqtSlot()
    def on_shiftBtn_clicked(self):
        try:
            const = float(self.shiftConstEdit.text())
        except ValueError:
            return
        self.data.x += const
        self.replotRequest.emit(None)
        session.set_dirty()

    @pyqtSlot()
    def on_monscaleBtn_clicked(self):
        try:
            const = int(self.monscaleEdit.text())
        except ValueError:
            return
        self.data.rescale(const)
        self.replotRequest.emit(None)
        session.set_dirty()

    @pyqtSlot()
    def on_titleBtn_clicked(self):
        self.data.meta.title = str(self.titleEdit.text())
        self.titleChanged.emit()
        session.set_dirty()
        self.replotRequest.emit(None)

    @pyqtSlot()
    def on_nameBtn_clicked(self):
        self.data.name = str(self.nameEdit.text())
        session.set_dirty()
        self.replotRequest.emit(None)

    @pyqtSlot()
    def on_subtractBtn_clicked(self):
        from ufit.gui.scanitem import ScanDataItem
        dlg = QDialog(self)
        loadUi(dlg, 'subtract.ui')
        data2obj = dlg.setList.populate(ScanDataItem)
        if dlg.exec() != QDialog.DialogCode.Accepted:
            return
        witems = dlg.setList.selectedItems()
        if not witems:
            return
        try:
            prec = float(dlg.precisionEdit.text())
        except ValueError:
            QMessageBox.warning(self, 'Error', 'Please enter a valid precision.')
            return

        new_data = self.data.subtract(data2obj[witems[0].type()].data,
                                      prec, dlg.destructBox.isChecked())

        if not dlg.destructBox.isChecked():
            new_model = self.model.copy()
            session.add_item(ScanDataItem(new_data, new_model), self.item.group)
        else:
            self.replotRequest.emit(None)
            session.set_dirty()

    @pyqtSlot()
    def on_fftBtn_clicked(self):
        try:
            npoints = int(self.fftNpointsEdit.text())
        except ValueError:
            QMessageBox.warning(self, 'Error',
                                'Please enter a valid number of points.')
            return
        xmin = self.data.x.min()
        xmax = self.data.x.max()
        xinterp = linspace(xmin, xmax, npoints)
        yinterp = interp1d(self.data.x, self.data.y, kind='linear')
        yfft = fft(yinterp(xinterp))
        p2 = abs(yfft) / npoints
        p1 = p2[:npoints//2 + 2]
        p1[1:-1] *= 2
        dx = (xmax - xmin) / (npoints - 1)

        new_data = ScanData.from_arrays(
            name='FFT(' + self.data.name + ')',
            x=(1./dx) * arange(npoints//2 + 2) / npoints,
            y=p1,
            dy=0.01*ones(p1.shape),
            xcol='1/' + self.data.xaxis,
            ycol='|P1|')
        new_model = self.model.copy()
        from ufit.gui.scanitem import ScanDataItem
        session.add_item(ScanDataItem(new_data, new_model), self.item.group)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for simple comma-separated column data files."""

# pylint: disable=unused-import
from ufit.data.simple import check_data_simple, guess_cols, read_data_simple


def read_data(filename, fp):
    return read_data_simple(filename, fp, ',')


def check_data(fp):
    return check_data_simple(fp, b',')



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

import numpy as np

import ufit.lab as uf


class QReader:
    """Read and parse the datafiles"""

    def __init__(self, template, numors, tolerance=0.01):
        """ The reader.
        template, numors - used by ufit read_numors for reading the data
        tolerance - maximum distance of points from plane to be included in reading
        """
        uf.set_datatemplate(template)
        try:
            self.datas = uf.read_numors(numors, 0)
        except Exception as err:
            print("Unexpected error:", err)
            self.datas = []

        self.tolerance = tolerance

    def get_points(self, v1, v2):
        """It will find points in read scan files which has hkle values
        These is then converted to (v1, v2) basis and calculated distance from it.
        Only inplane points are processed.
        """

        v = np.array([v1, v2, np.cross(v1, v2)]).transpose()
        pts = []
        for d in self.datas:
            # is it q-scan?
            if 'hkle' in d.meta:
                hkle = d.meta['hkle']
            else:
                print("File %s does not have hkle values for each point, "
                      "skipping." % d.name)
                continue

            for h, k, l, e in hkle:
                # transform coordinates to v1 and v2 basis
                mycoords = np.linalg.solve(v, [h, k, l])
                if abs(mycoords[2]) < self.tolerance:
                    pts.append([mycoords[0], mycoords[1], e])
                # else:
                #     print "out of plane reflection:", h, k, l

        return pts



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Load routine for NIST data files."""

import io
import re
import shlex

from numpy import array, concatenate, loadtxt, ones


def guess_cols(colnames, coldata, meta):
    mg = 'Mon'
    if 'hkle' in meta:
        xg = {'h': 'Q(x)', 'k': 'Q(y)', 'l': 'Q(z)', 'E': 'E'}[meta['hkle_vary']]
    else:
        xg = colnames[0]
    yg = None
    if 'Counts' in colnames:
        yg = 'Counts'
    elif 'COUNTS' in colnames:
        yg = 'COUNTS'
    return xg, yg, None, mg


def check_data(fp):
    fp.readline()
    line = fp.readline()
    fp.seek(0, 0)
    # on the first line is always name filename, on the second line
    # the description for it
    return line.startswith(b'  Filename')


def _maybe_float(text):
    try:
        return float(text)
    except ValueError:
        return text


def _read_meta(fp):
    # Read a single line of metadata, with the field names below
    values = shlex.split(fp.readline())
    fp.readline()
    return values


def read_data(filename, fp):
    fp = io.TextIOWrapper(fp, 'ascii', 'ignore')

    meta = {}
    meta['instrument'] = 'NIST'

    # metadata lines
    meta1 = _read_meta(fp)
    m = re.match(r'^(\D+)(\d+)', meta1[0])
    meta['experiment'] = m.group(1)
    meta['filenumber'] = int(m.group(2))
    meta['timestamp'] = meta1[1]
    meta['scantype'] = meta1[2]
    meta['title'] = fp.readline()

    meta2 = _read_meta(fp)
    meta['collimation'] = '-'.join(meta2[:4])
    meta['mosaic'] = '-'.join(meta2[4:7])
    meta['orientation'] = ' '.join(meta2[7:])

    meta3 = _read_meta(fp)
    meta['lattice'] = ' '.join(meta3[:3])
    meta['angles'] = ' '.join(meta3[3:])

    meta4 = _read_meta(fp)
    meta['temp'] = float(meta4[5])
    meta['delta_temp'] = float(meta4[6])

    meta5 = _read_meta(fp)
    meta['field'] = float(meta5[6])

    # principal axes
    if meta1[2] == 'B':
        mon = float(meta1[3])
        for ax in range(1, 7):
            meta['A%d' % ax] = float(fp.readline().split()[1])
        fp.readline()  # unimportant
    elif meta1[2] == 'Q':
        mon = float(meta1[3])
    else:
        raise Exception('unknown scan type %r' % meta1[2])

    colnames = fp.readline().split()
    arr = loadtxt(fp, ndmin=2)
    # if number of colnames is not correct, discard them
    if len(colnames) != arr.shape[1]:
        colnames = ['Column %d' % i for i in range(1, arr.shape[1]+1)]

    # synthesize monitor column
    colnames += ['Mon']
    arr = concatenate((arr, ones((len(arr), 1)) * mon), 1)

    cols = dict((name, arr[:, i]) for (i, name) in enumerate(colnames))
    meta['environment'] = []
    if 'Q(x)' in colnames:
        meta['hkle'] = arr[:, (0, 1, 2, 3)]
        deviations = array([(cs.max()-cs.min()) for cs in meta['hkle'].T])
        meta['hkle_vary'] = ['h', 'k', 'l', 'E'][deviations.argmax()]
    for col in cols:
        meta[col] = cols[col].mean()
    for tcol in ['TEMP', 'T-act']:
        if tcol in cols:
            meta['environment'].append('T = %.3f K' % meta[tcol])
            break

    return colnames, arr, meta



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

# for pickle compatibility
# pylint: disable=unused-import
from ufit.gui.scanitem import ScanDataItem as DatasetItem



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Data loading and treatment for ufit."""

from ufit import UFitError
from ufit.data import cascade, desy, dynacool, ill, llb, nicos, nicos_old, \
    nist, simple, simple_csv, taipan, trisp, zebra
from ufit.data.dataset import Dataset, DatasetList, ImageData, ScanData
from ufit.data.loader import Loader
from ufit.plotting import mapping

data_formats_scan = {
    'ill': ill,
    'nicos': nicos,
    'old nicos': nicos_old,
    'simple': simple,
    'simple comma-separated': simple_csv,
    'trisp': trisp,
    'llb': llb,
    'taipan': taipan,
    'nist': nist,
    'desy': desy,
    'zebra': zebra,
    'dynacool': dynacool,
}

data_formats_image = {
    'cascade': cascade,
}

data_formats = dict(list(data_formats_scan.items()) +
                    list(data_formats_image.items()))

__all__ = ['Dataset', 'DatasetList', 'ScanData', 'ImageData', 'sets',
           'set_datatemplate', 'set_dataformat', 'read_data', 'as_data',
           'read_numors', 'do_mapping']


# simplified interface for usage in noninteractive scripts

global_loader = Loader()
sets = global_loader.sets


def set_datatemplate(template):
    """Set a new template for data file names.

    Normally, data file names consist of a fixed part and a sequential number.
    Therefore ufit constructs file names from a data template, which should
    contain a placeholder like ``%06d`` (for a 6-digit sequential number), and
    the actual file number given in the :func:`read_data` function.

    An example::

       set_datatemplate('/data/exp/data2012n%06d.dat')
       d1 = read_data(100)
       d2 = read_data(101)
       # etc.
    """
    global_loader.template = template


def set_dataformat(fmt):
    """Set the input data format.

    Normally ufit autodetects file formats, but this can be overridden using
    this function.  Data formats are:

    * ``'ill'`` - ILL TAS and D23 data formats
    * ``'nicos'`` - NICOS 2.0+ data format
    * ``'old nicos'`` - NICOS 1.0 data format
    * ``'trisp'`` - FRM-II TRISP data format
    * ``'taipan'`` - ANSTO Taipan data format
    * ``'nist'`` - NIST data format
    * ``'llb'`` - LLB binary TAS data format (known working for data from
      1T and 4F unpolarized)
    * ``'desy'`` - DESY data format
    * ``'zebra'`` - PSI ZEBRA data format
    * ''`dynacool'`` - DynaCool PPMS format
    * ``'simple'`` - simple whitespace-separated multi-column files
    * ``'simple comma-separated'`` - simple comma-separated multi-column files
    """
    if fmt not in data_formats:
        raise UFitError('Unknown data format: %r, available formats are %s'
                        % (fmt, ', '.join(data_formats)))
    global_loader.format = fmt


# pylint: disable=redefined-builtin
def read_data(n, xcol='auto', ycol='auto', dycol=None, ncol=None, nscale=1, filter=None):
    """Read a data file.  Returns a :class:`Dataset` object.

    :param xcol: X column name (or 1-based index)
    :param ycol: Y column name (or 1-based index)
    :param ycol: Y errors column name (or 1-based index); the default is to take
        the square root of the Y column as appropriate for counts
    :param ncol: normalization column name (or 1-based index); typically a beam
        monitor column
    :param nscale: scale for the normalization column; the Y data is determined
        as ``y[i] = y_raw[i] / ncol[i] * nscale``
    :param filter: dictionary of column names and filtered values; only datapoints
        with given values will be part of the dataset; if None, then not applied.
    """
    return global_loader.load(n, xcol, ycol, dycol, ncol, nscale, filter)


def as_data(x, y, dy, name=''):
    """Quickly construct a :class:`Dataset` object from three numpy arrays."""
    return Dataset.from_arrays(name or 'data', x, y, dy)


def read_numors(nstring, binsize, xcol='auto', ycol='auto',
                dycol=None, ncol=None, nscale=1, floatmerge=True):
    """Read a number of data files.  Returns a list of :class:`Dataset`\\s.

    :param nstring: A string that gives file numbers, with the operators given
        below.
    :param binsize: Bin size when files need to be merged according to
        *nstring*.
    :param floatmerge: If to use float merging instead of binning. Default true.

    Other parameters as in :func:`read_data`.

    *nstring* can contain these operators:

    * ``,`` -- loads multiple files
    * ``-`` -- loads multiple sequential files
    * ``+`` -- merges multiple files
    * ``>`` -- merges multiple sequential files

    For example:

    * ``'10-15,23'`` loads files 10 through 15 and 23 in 7 separate datasets.
    * ``'10+11,23+24'`` loads two datasets consisting of files 10 and 11 merged
      into one set, as well as files 23 and 24.
    * ``'10>15+23'`` merges files 10 through 15 and 23 into one single dataset.
    * ``'10,11,12+13,14'`` loads four sets.
    """
    return global_loader.load_numors(nstring, binsize, xcol, ycol,
                                     dycol, ncol, nscale, floatmerge)


def do_mapping(x, y, runs, minmax=None, mode=0, log=False, dots=True,
               xscale=1, yscale=1, interpolate=100, usemask=True, figure=None,
               clear=True, colors=None, axes=None, title=None):
    """Create a 2D map from several datasets.

    An example::

        mapping("h", "E", datas)

    :param minmax: tuple of minimum and maximum value for z-scale
    :param mode: 0 = image, 1 = contour filled, 2 = contour lines
    :param log: True if the z-scale should be logarithmic
    :param dots: whether to show dots on position of datapoints
    :param xscale, yscale: to define scale between x and y axis
    :param interpolate: number of points to interpolate in between points
    :param usemask: True if masked out points should be removed
    :param figure: where to plot, default is current figure (``pl.gcf()``)
    :param clear: whether to do ``figure.clf()`` before plotting
    :param colors: when using mode > 1, optional list of colors for plotting
    :param axes: axes where to plot, default is figure.gca()
    :param title: title of the plot
    """
    return mapping(x, y, runs, minmax, mode, log, dots,
                   xscale, yscale, interpolate, usemask, figure,
                   clear, colors, axes, title)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Qt compatibility layer."""

# pylint: disable=wildcard-import, unused-import, unused-wildcard-import

import sys

try:
    import PyQt5

except ImportError:
    from PyQt6 import uic
    from PyQt6.QtCore import *
    from PyQt6.QtGui import *
    from PyQt6.QtPrintSupport import *
    from PyQt6.QtSvg import *
    from PyQt6.QtWidgets import *

    import ufit.guires_qt6

else:
    from PyQt5 import uic
    from PyQt5.QtCore import *
    from PyQt5.QtGui import *
    from PyQt5.QtPrintSupport import *
    from PyQt5.QtSvg import *
    from PyQt5.QtWidgets import *

    import ufit.guires_qt5

# Do not abort on exceptions in signal handlers.
# pylint: disable=unnecessary-lambda
sys.excepthook = lambda *args: sys.__excepthook__(*args)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

# pylint: disable=wrong-import-position, wrong-import-order

from ufit.version import get_version

__version__ = get_version()


class UFitError(Exception):
    pass

import matplotlib

try:
    import ufit.qt
    matplotlib.use('Qt5Agg')
except Exception:
    pass

from ufit.backends import *
from ufit.data import *
from ufit.param import *
from ufit.result import *

from ufit.models import *  # isort: skip



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Data merging routines."""

from numpy import append, arange, array, ones, reshape, sqrt, zeros

from ufit import UFitError


def rebin(data, binsize, meta=None):
    """Simple rebinning of (x, y, dy, n) data and col_ meta."""
    meta = meta or {}

    if binsize == 0:
        # no merging, just concatenate
        return data

    x, y, dy, n = data.T
    # copy meta
    new_meta = meta.copy()
    # identify columns
    metanames = []
    if meta != []:
        for col in meta:
            if not col.startswith('col_'):
                continue
            # add column to data
            metanames.append(col)

    # calculate new x values

    halfbinsize = binsize/2.
    stops = arange(x.min() - (x.min() % binsize) - binsize,
                   x.max() - (x.max() % binsize) + 2*binsize,
                   binsize) + halfbinsize
    nbins = len(stops)

    # newarray will be the new x, y, dy, n and meta columns array
    newarray = zeros((nbins, 4 + len(metanames)))
    newarray[:, 0] = stops

    # this will keep track which data values we already used
    data_unused = ones(len(x), bool)

    # this will keep track which new bins are used; unused ones are
    # left out
    new_used = ones(nbins, bool)

    for i in range(nbins):
        stop = newarray[i, 0]
        # get indices of all data points with x values lying below stop
        indices = x <= (stop + halfbinsize)
        # remove indices of data already used in previous bins
        indices &= data_unused
        if indices.any():
            newarray[i, 1] += y[indices].sum()
            newarray[i, 2] += sqrt((dy[indices]**2).sum())
            newarray[i, 3] += n[indices].sum()
            data_unused[indices] = False
            for j, m in enumerate(metanames):
                newarray[i, 4 + j] += meta[m][indices].sum() / sum(indices)
        else:
            new_used[i] = False
    # are there any data points left unused?
    if data_unused.any():
        raise UFitError('Merging data failed (data left over), check merging '
                        'algorithm for bugs')

    # remove any stops without monitor data
    newarray = newarray[new_used]

    # extract merged meta information
    for i, m in enumerate(metanames):
        new_meta[m] = array(newarray[:, 4 + i])
    # return arrays
    return array(newarray[:, :4]), new_meta


def mergeList(tomerge):
    tomerge = array(tomerge)
    merged = tomerge.sum(axis = 0)
    merged[2] = sqrt((tomerge**2).sum(axis = 0)[2])
    merged[0] = merged[0] / len(tomerge)
    # average meta
    merged[4:] = [x / len(tomerge) for x in merged[4:]]
    return merged


def floatmerge(data, binsize, meta=None):
    """Merging data based on floating window."""
    meta = meta or []

    if binsize == 0:
        # no merging, just concatenate
        return data, meta

    # sort data
    sortorder = data[:, 0].argsort()
    data = data[sortorder]
    # sort meta and empty new meta columns
    new_meta = meta.copy()
    metanames = []
    if meta:
        for col in meta:
            if not col.startswith('col_'):
                continue
            # add column to data
            metanames.append(col)
            data = append(data, reshape(meta.get(col, [])[sortorder], (-1, 1)),
                          axis=1)

    lastvals = []
    tomerge = []
    newlist = []

    for vals in data:
        # pylint: disable=use-implicit-booleaness-not-comparison
        if lastvals != []:  # no "if lastvals", could be an array
            if vals[0] > lastvals[0] + binsize:
                # merge points in list:
                newlist.append(mergeList(tomerge))
                tomerge = []
        tomerge.append(vals)
        lastvals = vals
    newlist.append(mergeList(tomerge))

    newlist = array(newlist)
    # extract merged meta information
    for i, m in enumerate(metanames):
        new_meta[m] = array(newlist[:, 4 + i])
    # return arrays
    return array(newlist[:, :4]), new_meta



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

from enum import Enum

import numpy as np

_marker_size = 10
_bz_width = 0.5
_label_size = 18
inva = np.array([-1, 1])


class PlaneType(Enum):
    HHLPLANE = 0
    H0LPLANE = 1
    HK0PLANE = 2

    def GetVector(inp):  # pylint: disable=no-self-argument
        if inp == PlaneType.H0LPLANE:
            return (0, 1, 0)
        if inp == PlaneType.HK0PLANE:
            return (0, 0, 1)
        if inp == PlaneType.HHLPLANE:
            return (1, -1, 0)
        raise NotImplementedError


class BZCreator:
    """ Experimantal version of the brillouin zone plotter """
    # Only BCT2 structure is supported now

    def __init__(self, gpts, a, c, plane):
        # just for bct2 - body centered tetragonal 2 brilluin zone yet
        if c < a:
            raise NotImplementedError
        # sp:
        self.a = a
        self.c = c

        self.sps = {
            "P":      [0.5,                     0.5,                     0.5],
            "Y1":     [(-a**2 + c**2)/(2*c**2), (-a**2 + c**2)/(2*c**2), 1.0],
            "Y":      [(a**2 + c**2)/(2*c**2),  (-a**2 + c**2)/(2*c**2), 0],
            "X":      [0.5,                     0.5,                     0],
            "Z":      [0,                       0,                       1.0],
            "Sigma":  [(a**2 + c**2)/(2*c**2),  0,                       0],
            "Sigma1": [(-a**2 + c**2)/(2*c**2), 0,                       1.0],
            "N":      [0.5,                     0,                       0.5],
            # "Gamma": [0, 0, 0],
        }
        self.gpts = gpts
        if plane[0] == plane[1] == 0:
            self.pt = PlaneType.HK0PLANE
        elif (plane[2] == 0) and (plane[0] == 0 or plane[1] == 0):
            self.pt = PlaneType.H0LPLANE
        elif (plane[2] == 0) and (abs(plane[1]) == abs(plane[0])):
            self.pt = PlaneType.HHLPLANE
        else:
            raise NotImplementedError

    def doPlot(self, plt, alph=0.1, show_labels=True, lw=_bz_width):

        plt.plot(self.gpts[:, 0], self.gpts[:, 1], 'o',
                 markersize=_marker_size + 3, alpha=alph)

        # pts
        pts = self.getPts()
        for p in pts:
            plt.plot(p[:, 0], p[:, 1], 'o',
                     markersize=_marker_size, alpha=alph)

        # lines
        pts = self.getLines()
        for p in pts:
            plt.plot(p[:, 0], p[:, 1], 'k--', linewidth=lw)

        # misc
        if self.pt == PlaneType.HK0PLANE:
            plt.set_aspect(1)
            plt.set_xlabel('h (r.l.u.)', fontsize=_label_size)
            plt.set_ylabel('k (r.l.u.)', fontsize=_label_size)
            if show_labels:
                plt.text(0.07, -0.05, r'$\Gamma$', fontsize=_label_size)
                plt.text(0.65, -0.05, r'$\Sigma$', fontsize=_label_size)
                plt.text(0.65,  0.27, r'Y', fontsize=_label_size)
                plt.text(0.55,  0.55, r'X', fontsize=_label_size)
        elif self.pt == PlaneType.H0LPLANE:
            plt.set_aspect(self.a/self.c)
            plt.set_xlabel('h (r.l.u.)', fontsize=_label_size)
            plt.set_ylabel('l (r.l.u.)', fontsize=_label_size)
            if show_labels:
                plt.text(0.07, -0.08, r'$\Gamma$', fontsize=_label_size)
                plt.text(0.45, -0.08, r'$\Sigma$', fontsize=_label_size)
                plt.text(0.57,  0.40, r'N', fontsize=_label_size)
                plt.text(0.07,  1.07, r'Z', fontsize=_label_size)
        elif self.pt == PlaneType.HHLPLANE:
            plt.set_aspect(self.a/self.c*np.sqrt(2))
            plt.set_xlabel('h (r.l.u.)', fontsize=_label_size)
            plt.set_ylabel('l (r.l.u.)', fontsize=_label_size)
            # if show_labels:
            #     plt.text(0.07, -0.08, r'$\Gamma$', fontsize=_label_size)
            #     plt.text(0.45, -0.08, r'$\Sigma$', fontsize=_label_size)
            #     plt.text(0.57,  0.40, r'N', fontsize=_label_size)
            #     plt.text(0.07,  1.07, r'Z', fontsize=_label_size)

    def getPts(self):

        retpts = []
        for sp in self.sps.values():
            # if (self.pt == PlaneType.H0LPLANE and sp[self.cz] == 0):  # only points in this plane
            if np.dot(PlaneType.GetVector(self.pt), sp) == 0:
                if self.pt in (PlaneType.H0LPLANE, PlaneType.HHLPLANE):
                    vec = [sp[0], sp[2]]
                elif self.pt == PlaneType.HK0PLANE:
                    vec = sp[:2]
                else:
                    raise NotImplementedError

                pts = []
                # print name, sp
                for gpt in self.gpts:
                    pts.append(gpt + vec)
                    pts.append(gpt - vec)
                    pts.append(gpt + vec * inva)
                    pts.append(gpt - vec * inva)
                    if self.pt == PlaneType.HK0PLANE:  # tetragonal
                        vec.reverse()
                        pts.append(gpt + vec)
                        pts.append(gpt - vec)
                        pts.append(gpt + vec * inva)
                        pts.append(gpt - vec * inva)
                retpts.append(np.array(pts))
        return retpts

    def getLines(self):

        def additem(col, main, v):
            col.append([main + v, np.arctan2(*(v))])

        retpts = []
        for gpt in self.gpts:
            pts = []
            for sp in self.sps.values():
                if np.dot(PlaneType.GetVector(self.pt), sp) == 0:  # only points in this plane
                    if self.pt in (PlaneType.H0LPLANE, PlaneType.HHLPLANE):
                        vec = [sp[0], sp[2]]
                    elif self.pt == PlaneType.HK0PLANE:
                        vec = sp[:2]
                    else:
                        raise NotImplementedError

                    additem(pts, gpt, vec)
                    vec[:] = [-x for x in vec]
                    additem(pts, gpt, vec)
                    vec = list(vec * inva)
                    additem(pts, gpt, vec)
                    vec[:] = [-x for x in vec]
                    additem(pts, gpt, vec)
                    if self.pt == PlaneType.HK0PLANE:  # tetragonal
                        vec.reverse()
                        additem(pts, gpt, vec)
                        vec[:] = [-x for x in vec]
                        additem(pts, gpt, vec)
                        vec = list(vec * inva)
                        additem(pts, gpt, vec)
                        vec[:] = [-x for x in vec]
                        additem(pts, gpt, vec)
            # pts.sort()
            # ptdict = {}
            # for pt in pts:
            #     ptdict[pt[1]] = pt[0]
            pts = sorted(pts, key=lambda s: s[1])

            pts = np.array(pts)[:, 0]
            pts = np.array(list(pts))
            retpts.append(pts)
        return retpts



import os
from os import path

from setuptools import setup


def find_packages():
    """Return a list of all ufit subpackages."""
    out = ['ufit']
    stack = [('ufit', 'ufit.')]
    while stack:
        where, prefix = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where, name)
            if '.' not in name and os.path.isdir(fn) and \
                    os.path.isfile(os.path.join(fn, '__init__.py')):
                out.append(prefix + name)
                stack.append((fn, prefix + name + '.'))
    return out


def find_ui_files():
    """Find all Qt .ui files in ufit.gui subpackages."""
    res = {}
    for root, dirs, files in os.walk('ufit/gui/ui'):
        uis = [uifile for uifile in files if uifile.endswith('.ui')]
        if uis:
            res[root.replace('/', '.')] = uis
    return res


ns = {'__file__': path.abspath(path.join(
    path.dirname(__file__), 'ufit', 'version.py'))}
exec(open("ufit/version.py").read(), ns)
version = ns['get_version']()

pkg_data = find_ui_files()
pkg_data.setdefault('ufit', []).append('RELEASE-VERSION')


setup(
    name='ufit',
    version=version,
    license='BSD-2-Clause',
    author='Georg Brandl',
    author_email='g.brandl@fz-juelich.de',
    description='Universal scattering data fitting tool',
    url='https://forge.frm2.tum.de/review/plugins/gitiles/mlz/general/ufit',
    packages=find_packages(),
    package_data=pkg_data,
    entry_points={
        'gui_scripts': [
            'ufitgui=ufit.gui:main',
        ],
    },
    install_requires=[
        'matplotlib>=3.0',
        'numpy',
        'scipy',
        'PyQt5',
        'lmfit',
    ],
    extras_require={
        'repl': 'qtconsole',
        'minuit': 'iminuit',
        'cluster': 'paramiko',
    },
)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""
Utilities for clustering execution of a piece of code to multiple hosts using
SSH transport.
"""

import hashlib
import pickle
import queue
import sys
import threading
from os import path
from time import time

import paramiko

keyname = path.expanduser('~/.ufitcluster/key')
clusterlist = []


def init_cluster():
    try:
        with open(path.expanduser('~/.ufitcluster/hosts'),
                  encoding='utf-8') as fp:
            for line in fp:  # user@host
                if line.startswith('#'):
                    continue
                login, host = line.strip().split('@')
                clusterlist.append((login, host))
    except Exception as err:
        print('Cannot read the list of cluster hosts:', err, file=sys.stderr)
        print('Please create a file %s with one line for each '
              'parallel process, for example:' %
              path.expanduser('~/.ufitcluster/hosts'), file=sys.stderr)
        print(file=sys.stderr)
        print('username@hostname1', file=sys.stderr)
        print('username@hostname1', file=sys.stderr)
        print('username@hostname1', file=sys.stderr)
        print('username@hostname1', file=sys.stderr)
        print('username@hostname2', file=sys.stderr)
        print('username@hostname2', file=sys.stderr)
        print(file=sys.stderr)
        print('would run 4 parallel processes on hostname1 and '
              '2 parallel processes on hostname2.', file=sys.stderr)
        print('For local operation only, use "localhost" as '
              'hostname.', file=sys.stderr)
        print('The file %s must contain the SSH private key to '
              'use for the connection, even for localhost.' %
              path.expanduser('~/.ufitcluster/key'), file=sys.stderr)


def client_runner(client, task_queue, result_queue):
    while True:
        code, funcname, jobnum, args = task_queue.get()
        if jobnum == -1:
            print('[C] exiting runner for %s' % client.ufit_host)
            return
        try:
            sidcontents = (str(time()) + str(args)).encode()
            sid = hashlib.md5(sidcontents).hexdigest()
            sys.stdout.write('.')
            sys.stdout.flush()
            # print '[C] starting job on %s: %s' % (client.ufit_host, sid)
            code_footer = '''\nif __name__ == "__main__":
            import pickle
            args = pickle.loads(%r)
            print pickle.dumps(%s(*args))
            ''' % (pickle.dumps(args), funcname)

            sftp = client.open_sftp()
            fobj = sftp.file('/tmp/ufit_cluster_%s.py' % sid, 'wb')
            fobj.write(code + code_footer)
            fobj.close()

            _stdin, stdout, _stderr = \
                client.exec_command('python /tmp/ufit_cluster_%s.py; '
                                    'rm /tmp/ufit_cluster_%s.py' % (sid, sid))
            result = pickle.load(stdout)
            # print '[C] done with job %s: %r' % (sid, result)
            result_queue.put((jobnum, result))
        except Exception as err:
            print(f'[C] no result on {client.ufit_host}, requeuing: {err!r}')
            task_queue.put((code, funcname, jobnum, args))
            return


clients = []
runners = []
global_task_queue = queue.Queue()
global_result_queue = queue.Queue()
cluster_setup = False


def setup_cluster():
    global cluster_setup
    if cluster_setup:
        return
    for i, (user, host) in enumerate(clusterlist):
        client = paramiko.SSHClient()
        client.ufit_host = '%s[%d]' % (host, i)
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(host, username=user, key_filename=keyname,
                       look_for_keys=False)
        clients.append(client)
        print(f'[C] opened cluster connection to {client.ufit_host}')
        runner = threading.Thread(target=client_runner,
                                  args=(client, global_task_queue,
                                        global_result_queue))
        runners.append(runner)
        runner.start()
    cluster_setup = True


def kill_cluster():
    for cl in clients:
        global_task_queue.put((None, None, -1, None))  # end!
        cl.close()


def run_cluster(code, funcname, argumentslist):
    setup_cluster()
    njobs = len(argumentslist)
    retval = [None] * njobs
    returns = 0
    for job in enumerate(argumentslist):
        global_task_queue.put((code, funcname) + job)
    while returns < njobs:
        # XXX check if any client is still running
        jobnum, result = global_result_queue.get()
        if isinstance(result, Exception):
            raise result
        retval[jobnum] = result
        returns += 1
    return retval


init_cluster()

if __name__ == '__main__':
    t1 = time()
    print(run_cluster('def foo(a): return a\n', 'foo', [('a',)]))
    t2 = time()
    print(t2-t1)



#  -*- coding: utf-8 -*-
# *****************************************************************************
# ufit, a universal scattering fitting suite
#
# Copyright (c) 2013-2023, Georg Brandl and contributors.  All rights reserved.
# Licensed under a 2-clause BSD license, see LICENSE.
# *****************************************************************************

"""Backend using iminuit."""

from iminuit import Minuit

from ufit.param import prepare_params, update_params
from ufit.utils import get_chisqr

__all__ = ['do_fit', 'backend_name']

backend_name = 'minuit'


def do_fit(data, fcn, params, add_kw):
    x, y, dy = data.fit_columns
    meta = data.meta
    varying, varynames, dependent, _ = prepare_params(params, meta)

    def minuitfcn(*args):
        pd = dict(zip(varynames, args))
        update_params(dependent, meta, pd)
        return ((fcn(pd, x) - y)**2 / dy**2).sum()

    print_report = add_kw.pop('printReport', False)

    marg = dict(
        print_level=int(print_report),
        forced_parameters=varynames,
    )
    marg.update(add_kw)
    if 'errordef' not in marg:
        marg['errordef'] = 1
    for p in varying:
        marg[p.name] = p.value
        marg['error_' + p.name] = 100*p.delta or p.value/100. or 0.01
        if p.pmin is not None or p.pmax is not None:
            marg['limit_' + p.name] = (p.pmin is None and -1e8 or p.pmin,
                                       p.pmax is None and +1e8 or p.pmax)

    m = Minuit(minuitfcn, **marg)
    try:
        m.migrad()
        m.hesse()
    except Exception as e:
        return False, str(e), 0
    # m.minos()  -> would calculate more exact and asymmetric errors

    pd = dict((pn, m.values[pn]) for pn in varynames)
    update_params(dependent, meta, pd)
    for p in params:
        p.value = pd[p.name]
        if p.name in varynames:
            p.error = m.errors[p.name]
            p.correl = {}  # XXX
        else:
            p.error = 0
            p.correl = {}

    return True, '', get_chisqr(fcn, x, y, dy, params)
